import json

from collections import namedtuple
from typing import Any, List, Tuple

from donphan import MaybeAcquire

from .tables import ALL_TABLES
from .utils import get_base_dir


BASE_DIR = get_base_dir()


async def setup_ampharos():
    """Populates the Pokemon database.

    This method should always be called on startup"""
    async with MaybeAcquire() as connection:

        # Populate the tables if required
        for table in ALL_TABLES:

            # If Table is empty
            if (await table.fetchrow(connection=connection)) is None:

                record = namedtuple("record", (column.name for column in table._columns))
                data: List[Tuple[Any, ...]] = []

                try:
                    with open(BASE_DIR / f"data/{table.__name__.lower()}.json") as f:
                        for item in json.load(f):
                            data.append(record(**item))
                except FileNotFoundError:
                    print(f"Could not find Pokemon data file {table.__name__.lower()}.json")
                else:
                    try:
                        await table.insert_many(table._columns, *data, connection=connection)
                    except Exception as e:
                        print(e)



from dataclasses import dataclass
from enum import Enum
from typing import List


class Category(Enum):
    """An enumeration of Pokemon :class:`types.Move` categories."""

    PHYSICAL = 1
    SPECIAL = 2
    STATUS = 3
    VARIES = 4


@dataclass
class _BasePokemonObject:
    _term: str


@dataclass
class Ability(_BasePokemonObject):
    """Represents an ability.

    Attributes:
        name (str): The ability's name
        description (str): A description of the ability
        introduced (int): The generation the ability was introduced
    """

    name: str
    description: str
    introduced: int


@dataclass
class Item(_BasePokemonObject):
    """Represents an item.

    Attributes:
        name (str): The item's name
        description (str): A description of the item
    """

    name: str
    description: str


@dataclass
class Typing(_BasePokemonObject):
    """Represents a Typing.

    Attributes:
        name (str): The typing's name
    """

    name: str


@dataclass
class Move(_BasePokemonObject):
    """Represents a move.

    Attributes:
        type (types.Typing): The move's typing
        name (str): The move's name
        description (str): A description of the move
        pp (int): The move's PP
        power (int): The move's power (`None` if status move or varies)
        accuracy (int): The move's accuracy (`None` if move cannot miss)
        category (types.Category): The moves category
    """

    type: Typing
    name: str
    description: str
    pp: int
    power: int
    accuracy: int
    category: Category


@dataclass
class PokemonName(_BasePokemonObject):
    """Represents a Pokemon's name.

    Attributes:
        english (str): The name in English
        japanese (str): The name in Romanji
        kana (str): The name in Kana
    """

    english: str
    japanese: str
    kana: str


@dataclass
class PokemonBaseStats(_BasePokemonObject):
    """Represents a Pokemon's base stats.

    Attributes:
        hp (int)
        attack (int)
        defense (int)
        special_attack (int)
        special_defense (int)
        speed (int)
        total (int)
    """

    hp: int
    attack: int
    defense: int
    special_attack: int
    special_defense: int
    speed: int

    @property
    def total(self):
        return sum([self.hp, self.attack, self.defense, self.special_attack, self.special_defense, self.speed])


@dataclass
class PokemonTypings(_BasePokemonObject):
    """Represents a Pokemon's Typing.

    Attributes:
        primary (types.Typing)
        secondary (types.Typing)
    """

    primary: Typing
    secondary: Typing


@dataclass
class PokemonPokedexEntries(_BasePokemonObject):
    """Represents a Pokemon's Pokedex Entries.

    Attributes:
        sun (str): Pokedex entry in Pokemon Sun.
        moon (str): Pokedex entry in Pokemon Moon.
    """

    sun: str
    moon: str


@dataclass
class PokemonAbilities(_BasePokemonObject):
    """Represents a Pokemon's Abilities.

    Attributes:
        primary (types.Ability)
        secondary (types.Ability)
        hidden (types.Ability)
    """

    primary: Ability
    secondary: Ability
    hidden: Ability


@dataclass()
class Pokemon(_BasePokemonObject):
    """Represents a Pokemon.

    Attributes:
        name (types.PokemonName): The Pokemon's Name
        pokedex_number (int): The Pokemon's Pokedex Number
        classification (str): The Pokemon's classification
        typing (types.PokemonTypings): The Pokemon's Typing
        pokedex_entries (types.PokemonPokedexEntries): The Pokemon's Pokedex Entries
        evolutions (List[types.Pokemon]): A list of Pokemon this Pokemon can evolve into
        base_stats (types.PokemonBaseStats): The Pokemon's Base Stats
        abilities: (types.PokemonAbilities): The Pokemon's Abilities

    """

    pokedex_number: int
    classification: str
    name: PokemonName
    pokedex_entries: PokemonPokedexEntries
    evolutions: List["Pokemon"]
    base_stats: PokemonBaseStats
    typing: PokemonTypings
    abilities: PokemonAbilities



__title__ = "ampharos"
__author__ = "Bijij"
__license__ = "MIT"
__copyright__ = "Copyright 2019-present Bijij"
__version__ = "2.0.1"

from . import tables as tables
from .search import *
from .setup import setup_ampharos as setup_ampharos



import difflib
import random

from typing import AsyncIterator, Optional, Type, TypeVar

import asyncpg

from donphan import Table, MaybeAcquire

from . import types, tables

__all__ = (
    "ability",
    "item",
    "move",
    "pokemon",
    "typing",
    "all_abilities",
    "all_items",
    "all_moves",
    "all_pokemon",
    "random_ability",
    "random_item",
    "random_move",
    "random_pokemon",
)

T = TypeVar("T")


async def _search(
    table: Type[Table],
    search_term: str,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[asyncpg.Record]:
    async with MaybeAcquire(connection) as connection:
        records = await connection.fetch(f"SELECT term FROM {table._name}")
        matches = difflib.get_close_matches(search_term, (record["term"] for record in records))

        if not matches:
            return None

        return await table.fetchrow(term=matches[0], connection=connection)


async def _fetch(
    table: Type[Table],
    term: str,
    type: Type[T],
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[T]:
    record = await table.fetchrow(connection=connection, term=term)
    return type(*record.values())


async def _ability(
    record: asyncpg.Record,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> types.Ability:
    return types.Ability(*record.values())


async def ability(
    search_term: str,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[types.Ability]:
    """Searches for a :class:`types.Ability`.

    Args:
        search_term (str): The term to search for
    Returns:
        types.Ability: The best matching ability.

    """
    async with MaybeAcquire(connection) as connection:
        record = await _search(tables.Abilities, search_term, connection=connection)
        if record is None:
            return None

        return await _ability(record, connection=connection)


async def _item(
    record: asyncpg.Record,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> types.Item:
    return types.Item(*record.values())


async def item(
    search_term: str,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[types.Item]:
    """Searches for a :class:`types.Item`.

    Args:
        search_term (str): The term to search for
    Returns:
        types.Item: The best matching item.
    """
    async with MaybeAcquire(connection) as connection:
        record = await _search(tables.Items, search_term, connection=connection)
        if record is None:
            return None

        return await _item(record, connection=connection)


async def _move(
    record: asyncpg.Record,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> types.Move:
    dct = dict(record.items())
    dct["type"] = await _fetch(tables.Types, dct["type"], types.Typing, connection=connection)
    dct["category"] = types.Category(dct["category"])

    return types.Move(*dct.values())


async def move(
    search_term: str,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[types.Move]:
    """Searches for a :class:`types.Move`.

    Args:
        search_term (str): The term to search for
    Returns:
        types.Move: The best matching move.
    """
    async with MaybeAcquire(connection) as connection:
        record = await _search(tables.Moves, search_term, connection=connection)
        if record is None:
            return None

        return await _move(record, connection=connection)


async def _pokemon(
    record: asyncpg.Record,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> types.Pokemon:
    async with MaybeAcquire(connection) as connection:
        dct = dict(record.items())
        term = dct["term"]

        # Retrieve additional objects
        dct["name"] = await _fetch(tables.PokemonNames, term, types.PokemonName, connection=connection)
        dct["pokedex_entries"] = await _fetch(
            tables.PokemonDexEntries, term, types.PokemonPokedexEntries, connection=connection
        )

        evolutions = await tables.PokemonEvolutions.fetch(connection=connection, term=term)
        dct["evolutions"] = [await pokemon(record["evolution"], connection=connection) for record in evolutions]

        dct["base_stats"] = await _fetch(tables.PokemonBaseStats, term, types.PokemonBaseStats, connection=connection)

        typing_dct = dict((await tables.PokemonTypes.fetchrow(connection=connection, term=term)).items())
        for key, value in typing_dct.items():
            if key == "term" or value is None:
                continue
            typing_dct[key] = await _fetch(tables.Types, typing_dct[key], types.Typing, connection=connection)
        dct["typing"] = types.PokemonTypings(*typing_dct.values())

        abilities_dct = dict((await tables.PokemonAbilities.fetchrow(connection=connection, term=term)).items())
        for key, value in abilities_dct.items():
            if key == "term" or value is None:
                continue
            abilities_dct[key] = await ability(abilities_dct[key], connection=connection)
        dct["abilities"] = types.PokemonAbilities(*abilities_dct.values())

        return types.Pokemon(*dct.values())


async def pokemon(
    search_term: str,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> Optional[types.Pokemon]:
    """Searches for a :class:`types.Pokemon`.

    Args:
        search_term (str): The term to search for
    Returns:
        types.Pokemon: The best matching Pokemon.
    """
    async with MaybeAcquire(connection) as connection:
        if search_term.isdigit():
            record = await tables.Pokemon.fetchrow(connection=connection, dex_no=int(search_term))
        else:
            record = await _search(tables.Pokemon, search_term, connection=connection)

        if record is None:
            return None

        return await _pokemon(record, connection=connection)


async def _typing(
    record: asyncpg.Record,
    *,
    connection: Optional[asyncpg.Connection] = None,
) -> types.Typing:
    return types.Typing(*record.values())


async def typing(search_term: str, *, connection: Optional[asyncpg.Connection] = None) -> Optional[types.Typing]:
    """Searches for a :class:`types.Typing`.

    Args:
        search_term (str): The term to search for
    """
    async with MaybeAcquire(connection) as connection:
        record = await _search(tables.Types, search_term, connection=connection)
        if record is None:
            return None

        return await _typing(record, connection=connection)


async def all_abilities(*, connection: Optional[asyncpg.Connection] = None) -> AsyncIterator[types.Ability]:
    """Returns an :class:`AsyncGenerator` of all :class:`types.Ability` in the database."""
    async with MaybeAcquire(connection) as connection:
        for record in await tables.Abilities.fetchall(connection=connection):
            yield await _ability(record, connection=connection)


async def all_items(*, connection: Optional[asyncpg.Connection] = None) -> AsyncIterator[types.Item]:
    """Returns an :class:`AsyncGenerator` of all :class:`types.Item` in the database."""
    async with MaybeAcquire(connection) as connection:
        for record in await tables.Items.fetchall(connection=connection):
            yield await _item(record, connection=connection)


async def all_moves(*, connection: Optional[asyncpg.Connection] = None) -> AsyncIterator[types.Move]:
    """Returns an :class:`AsyncGenerator` of all :class:`types.Move` in the database."""
    async with MaybeAcquire() as connection:
        for record in await tables.Moves.fetchall(connection=connection):
            yield await _move(record, connection=connection)


async def all_pokemon(*, connection: Optional[asyncpg.Connection] = None) -> AsyncIterator[types.Pokemon]:
    """Returns an :class:`AsyncGenerator` of all :class:`types.Pokemon` in the database."""
    async with MaybeAcquire() as connection:
        for record in await tables.Pokemon.fetchall(connection=connection):
            yield await _pokemon(record, connection=connection)


async def random_ability(*, connection: Optional[asyncpg.Connection] = None) -> types.Ability:
    """Returns a random :class:`types.Ability`."""
    async with MaybeAcquire(connection) as connection:
        records = await tables.Abilities.fetchall(connection=connection)

        return await _ability(random.choice(records), connection=connection)


async def random_item(*, connection: Optional[asyncpg.Connection] = None) -> types.Item:
    """Returns a random :class:`types.Item`."""
    async with MaybeAcquire(connection) as connection:
        records = await tables.Items.fetchall(connection=connection)

        return await _item(random.choice(records), connection=connection)


async def random_move(*, connection: Optional[asyncpg.Connection] = None) -> types.Move:
    """Returns a random :class:`types.Move`."""
    async with MaybeAcquire(connection) as connection:
        records = await tables.Moves.fetchall(connection=connection)

        return await _move(random.choice(records), connection=connection)


async def random_pokemon(*, connection: Optional[asyncpg.Connection] = None) -> types.Pokemon:
    """Returns a random :class:`types.Pokemon`."""
    async with MaybeAcquire(connection) as connection:
        records = await tables.Pokemon.fetchall(connection=connection)

        return await _pokemon(random.choice(records), connection=connection)



from typing import List, Type

from donphan import Table, SQLType, Column


class Abilities(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True)
    name: SQLType.Text
    description: SQLType.Text
    introduced: SQLType.Integer


class Items(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True)
    name: SQLType.Text
    description: SQLType.Text


class Types(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True)
    name: SQLType.Text


class Moves(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True)
    type: SQLType.Text = Column(references=Types.term)
    name: SQLType.Text
    description: SQLType.Text
    pp: SQLType.SmallInt
    power: SQLType.SmallInt
    accuracy: SQLType.SmallInt
    category: SQLType.SmallInt


class Pokemon(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True)
    dex_no: SQLType.SmallInt
    classification: SQLType.Text


class PokemonNames(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    english: SQLType.Text
    japanese: SQLType.Text
    kana: SQLType.Text


class PokemonEvolutions(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    evolution: SQLType.Text = Column(primary_key=True, references=Pokemon.term)


class PokemonDexEntries(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    sun: SQLType.Text
    moon: SQLType.Text


class PokemonTypes(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    first: SQLType.Text = Column(references=Types.term)
    second: SQLType.Text = Column(references=Types.term)


class PokemonAbilities(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    first: SQLType.Text = Column(references=Abilities.term)
    second: SQLType.Text = Column(references=Abilities.term)
    hidden: SQLType.Text = Column(references=Abilities.term)


class PokemonBaseStats(Table, schema="ampharos"):  # type: ignore[call-arg]
    term: SQLType.Text = Column(primary_key=True, references=Pokemon.term)
    hp: SQLType.SmallInt
    attack: SQLType.SmallInt
    defense: SQLType.SmallInt
    special_attack: SQLType.SmallInt
    special_defense: SQLType.SmallInt
    speed: SQLType.SmallInt


ALL_TABLES: List[Type[Table]] = [
    Abilities,
    Items,
    Types,
    Moves,
    Pokemon,
    PokemonEvolutions,
    PokemonDexEntries,
    PokemonNames,
    PokemonTypes,
    PokemonAbilities,
    PokemonBaseStats,
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ampharos']

package_data = \
{'': ['*'], 'ampharos': ['data/*']}

install_requires = \
['donphan>=3.4']

setup_kwargs = {
    'name': 'ampharos',
    'version': '2.0.1',
    'description': 'A Pokemon Database for Donphan',
    'long_description': 'Ampharos\n========\n\nLightweight Pokemon Database for use with `Donphan <https://github.com/bijij/donphan>`_ ORM.\n',
    'author': 'bijij',
    'author_email': 'josh@josh-is.gay',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.2,<4.0.0',
}


setup(**setup_kwargs)
